/*
 * kdesupport.c - KDE support routines. Works without KDE, too.
 *
 * (c) 1998 by Martin Denn <Martin.Denn@rhein-main.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "kdesupport.h"

#define MAXDEPTH 20

#define UNINITIALIZED -1
#define KDE_NOT_DETECTED 0
#define KDE_DETECTED 1

typedef struct _LookupTable LookupTable;
struct _LookupTable {
    char *key, *value;
    LookupTable *left, *right;
};

static String _KDE_resource_translations = NULL;
static Widget _KDE_top_level_widget;
static Window _KDE_change_palette_window;
static Atom _KDE_change_palette;
static int _KDE_running = UNINITIALIZED;

/* implement simple lookup table as binary tree */
void insertInTable(LookupTable **root, char *key, char *value) {
    int cmp_result;

    if ((*root) == (LookupTable *) NULL) {
        (*root) = (LookupTable *) malloc(sizeof(LookupTable));
        (*root)->key = (char *) malloc(strlen(key)+1);
        (*root)->value = (char *) malloc(strlen(value)+1);
        strcpy((*root)->key, key);
        strcpy((*root)->value, value);
        (*root)->left = (LookupTable *) NULL;
        (*root)->right = (LookupTable *) NULL;
    }
    else {
        cmp_result = strcmp(key, (*root)->key);
        if (cmp_result < 0) {
            insertInTable(&((*root)->left), key, value);
        }
        else if (cmp_result > 0) {
            insertInTable(&((*root)->right), key, value);
        }
        else {
            free((*root)->value);
            (*root)->value = (char *) malloc(strlen(value)+1);
            strcpy((*root)->value, value);
        }
    }
}
            
void deleteTable(LookupTable **root) {
    if ((*root) != (LookupTable *) NULL) {
        deleteTable(&((*root)->left));
        deleteTable(&((*root)->right));
        free((*root)->key);
        free((*root)->value);
        free(*root);
    }
}

char *findInTable(LookupTable *root, char *key) {
    int cmp_result;
    
    if (root != (LookupTable *) NULL) {
        cmp_result = strcmp(key, root->key);
        if (cmp_result < 0) {
            return findInTable(root->left, key);
        }
        else if (cmp_result > 0) {
            return findInTable(root->right, key);
        }
        else {
            return root->value;
        }
    }
    else 
        return NULL;
}

#ifdef DEBUG
void printTable(LookupTable *root) {
    if (root != (LookupTable *) NULL) {
        printTable(root->left);
        printf("%s:%s\n", root->key, root->value);
        printTable(root->right);
    }
}
#endif


/* read $HOME/.kderc and fill lookup table with the needed values */
void readKDErc(LookupTable **root) {
    char *home, *kdeconfig, *cp, s[256];
    FILE *fd;
    int valid;
#ifndef KDE_BETA_3
    int red, green, blue;
#endif

    deleteTable(root);
    home=getenv("HOME");
    kdeconfig=malloc(strlen(home)+8);
    strcpy(kdeconfig, home);
    strcat(kdeconfig, "/.kderc");
    fd=fopen(kdeconfig, "r");
    if (fd!=NULL) {
        valid=0;
        while (fgets(s, 256, fd)!=NULL) {
            if (s[0]=='[') {
#ifdef KDE_BETA_3
                valid=(strncmp(s, "[Color Scheme]", 14)==0);
#else
                valid=(strncmp(s, "[General]", 9)==0);
#endif
            } 
            else if (valid) {
                s[strlen(s)-1]='\0';
                cp=strchr(s, '=');
                if (cp==NULL) cp=s+strlen(s)-1;
                else (*cp)='\0';
#ifndef KDE_BETA_3
                sscanf(cp+1, "%d,%d,%d", &red, &green, &blue);
                sprintf(cp+1, "#%02x%02x%02x", red, green, blue);
#endif
                insertInTable(root, s, cp+1);
            }
        }
        fclose(fd);
    }
    /* printTable(*root); */
}

void buildKDEResourceDB(XrmDatabase *db, LookupTable *root, 
                        char *kderesource) {
    char *cp, *resource, *key, *value, *db_line, *s;

    s = (char *) malloc(strlen(kderesource)+1);
    strcpy(s, kderesource);
    cp = s;
    while ((*cp) != '\0') {
        while (isspace(*cp)) cp++;
        resource=cp; 
        while (((*cp) != ':') && ((*cp) != '\0')) cp++;
        if ((*cp) == '\0') {
            fprintf(stderr, "KDE resource conversion: Error parsing %s\n",
                    resource);
        }
        else {
            (*cp)='\0'; cp++;
            while (isspace(*cp)) cp++; 
            key=cp;
            while (isalnum(*cp)) cp++;
            if ((*cp)!='\0') {
                (*cp)='\0'; cp++;
            }
            value=findInTable(root, key);
            if (value==NULL) value=key;
            db_line = (char *) malloc(strlen(resource)+strlen(value)+2);
            strcpy(db_line, resource);
            strcat(db_line, ":");
            strcat(db_line, value);
            XrmPutLineResource(db, db_line);
#ifdef DEBUG
            printf("%s\n", db_line);
#endif
        }
    }
    free(s);
}
    
void setResources(Widget w, XrmDatabase db,
               XrmNameList name, XrmClassList class, int depth);

void subtree(WidgetList w_list, Cardinal w_num, XrmDatabase db, 
             XrmNameList name, XrmClassList class, int depth) {
    unsigned int i;

    if (depth+1<MAXDEPTH) {
        for (i=0; i<w_num; i++) {
            if (w_list[i] != NULL) {
                name[depth] = w_list[i]->core.xrm_name;
                class[depth] = w_list[i]->core.widget_class->
                    core_class.xrm_class;
                setResources(w_list[i], db, name, class, depth+1);
            }
        }
    }
}

void setResources(Widget w, XrmDatabase db, 
               XrmNameList name, XrmClassList class, int depth) {
    XtResourceList res;
    XrmRepresentation representation;
    XrmValue from, to;
    XtArgVal *value;
    Cardinal num_res, w_num;
    WidgetList w_list;
    int i;

    /* get widget resources */
    XtGetResourceList(XtClass(w), &res, &num_res);
    name[depth+1] = NULLQUARK;
    class[depth+1]= NULLQUARK;
    for (i=0; i<num_res; i++) {
        name[depth] = XrmStringToQuark(res[i].resource_name);
        class[depth] = XrmStringToQuark(res[i].resource_class);
        if (XrmQGetResource(db, name, class, &representation, &from)==True) {
            /* allocate memory */
            value = (XtArgVal *) XtMalloc(res[i].resource_size);
            to.size = res[i].resource_size;
            to.addr = (XPointer) value;
            /* convert */
            if (XtConvertAndStore(w, XrmQuarkToString(representation), 
                                  &from, res[i].resource_type, &to)==False) {
                XtAppError(XtWidgetToApplicationContext(w), 
                           "conversion failed");
            }
            /* set */
            else {
                XtVaSetValues(w, res[i].resource_name, (*value), NULL);
            }
            /* free memory */
            XtFree((XPointer) value);
        }
    }
    XtFree((XPointer) res);
    /* now find out about child widgets. Here we access the deeper
       internals of Xlib. Be careful ! */ 
    if (XtIsComposite(w)) {
        w_num = ((CompositeRec *)w)->composite.num_children;
        w_list = ((CompositeRec *)w)->composite.children;
        if (w_num>0) subtree(w_list, w_num, db, name, class, depth);
    }
    if (XtIsShell(w)) {
        w_num = w->core.num_popups;
        w_list = w->core.popup_list;
        if (w_num>0) subtree(w_list, w_num, db, name, class, depth);
    }
}

void refreshResources() {
    String app_name, app_class;
    XrmQuark name[MAXDEPTH], class[MAXDEPTH];
    XrmDatabase db, app_db;
    LookupTable *l=NULL;

    if (_KDE_running == KDE_DETECTED) {

        /* read and store .kderc */
        readKDErc(&l);

        /* build database */
        db=NULL;
        buildKDEResourceDB(&db, l, _KDE_resource_translations);

        /* go down the tree and set resources */
        XtGetApplicationNameAndClass(XtDisplay(_KDE_top_level_widget), 
                                     &app_name, &app_class);
        name[0]=XrmStringToQuark(app_name);
        class[0]=XrmStringToQuark(app_class);
        setResources(_KDE_top_level_widget, db, name, class, 1);

        /* merge database into the official database */
        app_db=XtDatabase(XtDisplay(_KDE_top_level_widget));
        XrmMergeDatabases(db, &app_db);
    }
}

Boolean detectKDE(Widget w) {
    Atom a, type;
    int format;
    unsigned char *value;
    unsigned long bytes_ret, no_items;
 
    _KDE_running = KDE_NOT_DETECTED;
    if ((a = XInternAtom(XtDisplay(w), "KWM_RUNNING", True)) != None) {
        XGetWindowProperty(XtDisplay(w), RootWindowOfScreen(XtScreen(w)), 
                           a, 0, 1, False, AnyPropertyType, &type, &format,
                           &no_items, &bytes_ret, &value);
        if ((type != None) && (format == 32) && (no_items>0) && 
            ((long) (*value) == 1))
            _KDE_running = KDE_DETECTED;
        if (bytes_ret>0) XFree(value);
    }
    return (_KDE_running == KDE_DETECTED);
}

void registerResourceTranslation(Widget w, String s) {
    Atom a;
    unsigned long data[1];

    _KDE_resource_translations = s;
    _KDE_top_level_widget = w;
    if (_KDE_running == UNINITIALIZED) 
        detectKDE(w);
    if (_KDE_running == KDE_DETECTED) {
        _KDE_change_palette = XInternAtom(XtDisplay(w), 
                                          "KDEChangePalette", False);
        a = XInternAtom(XtDisplay(w), "KDE_DESKTOP_WINDOW", False);
        /* register desktop color change protocol.
           this protocol requires a simple window which is a direct
           child of the root window. So we create it using a call
           from Xlib. Xt does not know anything about this window, 
           so we have to process its events seperately */
        _KDE_change_palette_window = XCreateSimpleWindow(XtDisplay(w), 
                  RootWindowOfScreen(XtScreen(w)), 0, 0, 1, 1, 0, 0, 0);
        data[0] = 1;
        XChangeProperty(XtDisplay(w), _KDE_change_palette_window, a, a, 
                        32, PropModeReplace, (unsigned char *) data, 1);
        refreshResources();
    }
}

void processKDEEvents(XEvent ev) {
    /* special event processing */
    if ((_KDE_running == KDE_DETECTED) &&
        (ev.xany.type == ClientMessage) &&
        (ev.xclient.message_type == _KDE_change_palette)) {
        refreshResources();
    }
}
    

void registerMiniIcon(Widget w, Pixmap pixmap, Pixmap mask) {
    Atom a;
    unsigned long data[2];
    int i;

    if (_KDE_running == UNINITIALIZED) 
        detectKDE(w);
    if (_KDE_running == KDE_DETECTED) {
        a = XInternAtom(XtDisplay(w), "KWM_WIN_ICON", False);
        i=0; data[i++] = (long) pixmap;
        if (mask != XtUnspecifiedPixmap)
            data[i++] = (long) mask;
        XChangeProperty(XtDisplay(w), XtWindow(w), a, a, 32,
                        PropModeReplace, (unsigned char *) data, i);
    }
}

